@@ -1,5 +1,5 @@
 package Net::Async::AMQP::Channel;
-$Net::Async::AMQP::Channel::VERSION = '0.012';
+$Net::Async::AMQP::Channel::VERSION = '0.013';
 use strict;
 use warnings;
 
@@ -11,7 +11,7 @@ Net::Async::AMQP::Channel - represents a single channel in an MQ connection
 
 =head1 VERSION
 
-version 0.012
+version 0.013
 
 =head1 SYNOPSIS
 
@@ -52,13 +52,13 @@ use Data::Dumper;
 use Scalar::Util qw(weaken);
 
 use Net::Async::AMQP;
-use constant DEBUG => Net::Async::AMQP->DEBUG;
+use Net::Async::AMQP::Utils;
 
 use overload
-    '""' => sub { shift->as_string },
-    '0+' => sub { 0 + shift->id },
-    bool => sub { 1 },
-    fallback => 1;
+	'""' => sub { shift->as_string },
+	'0+' => sub { 0 + shift->id },
+	bool => sub { 1 },
+	fallback => 1;
 
 =head1 METHODS
 
@@ -72,7 +72,7 @@ sub configure {
 	for(grep exists $args{$_}, qw(future id)) {
 		$self->{$_} = delete $args{$_};
 	}
-    $self->SUPER::configure(%args);
+	$self->SUPER::configure(%args);
 }
 
 =head2 confirm_mode
@@ -94,22 +94,22 @@ channel once complete.
 =cut
 
 sub confirm_mode {
-    my $self = shift;
-    my %args = @_;
-    $self->debug_printf("Enabling confirm mode");
+	my $self = shift;
+	my %args = @_;
+	$self->debug_printf("Enabling confirm mode");
 
-    my $f = $self->loop->new_future;
-    my $frame = Net::AMQP::Frame::Method->new(
-        method_frame => Net::AMQP::Protocol::Confirm::Select->new(
-            nowait      => 0,
-        )
-    );
-    $self->push_pending(
-        'Confirm::SelectOk' => [ $f, $self ]
-    );
+	my $f = $self->loop->new_future;
+	my $frame = Net::AMQP::Frame::Method->new(
+		method_frame => Net::AMQP::Protocol::Confirm::Select->new(
+			nowait      => 0,
+		)
+	);
+	$self->push_pending(
+		'Confirm::SelectOk' => [ $f, $self ]
+	);
 	$self->closure_protection($f);
-    $self->send_frame($frame);
-    return $f;
+	$self->send_frame($frame);
+	return $f;
 }
 
 =head2 exchange_declare
@@ -128,32 +128,32 @@ channel once complete.
 =cut
 
 sub exchange_declare {
-    my $self = shift;
-    my %args = @_;
-    die "No exchange specified" unless exists $args{exchange};
-    die "No exchange type specified" unless exists $args{type};
+	my $self = shift;
+	my %args = @_;
+	die "No exchange specified" unless exists $args{exchange};
+	die "No exchange type specified" unless exists $args{type};
 
-    $self->debug_printf("Declaring exchange [%s]", $args{exchange});
+	$self->debug_printf("Declaring exchange [%s]", $args{exchange});
 
-    my $f = $self->loop->new_future;
-    my $frame = Net::AMQP::Frame::Method->new(
-        method_frame => Net::AMQP::Protocol::Exchange::Declare->new(
-            exchange    => Net::AMQP::Value::String->new($args{exchange}),
-            type        => Net::AMQP::Value::String->new($args{type}),
-            passive     => $args{passive} || 0,
-            durable     => $args{durable} || 0,
-            auto_delete => $args{auto_delete} || 0,
-            internal    => $args{internal} || 0,
-            ticket      => 0,
-            nowait      => 0,
-        )
-    );
-    $self->push_pending(
-        'Exchange::DeclareOk' => [ $f, $self ]
-    );
+	my $f = $self->loop->new_future;
+	my $frame = Net::AMQP::Frame::Method->new(
+		method_frame => Net::AMQP::Protocol::Exchange::Declare->new(
+			exchange    => Net::AMQP::Value::String->new($args{exchange}),
+			type        => Net::AMQP::Value::String->new($args{type}),
+			passive     => $args{passive} || 0,
+			durable     => $args{durable} || 0,
+			auto_delete => $args{auto_delete} || 0,
+			internal    => $args{internal} || 0,
+			ticket      => 0,
+			nowait      => 0,
+		)
+	);
+	$self->push_pending(
+		'Exchange::DeclareOk' => [ $f, $self ]
+	);
 	$self->closure_protection($f);
-    $self->send_frame($frame);
-    return $f;
+	$self->send_frame($frame);
+	return $f;
 }
 
 =head2 queue_declare
@@ -168,13 +168,13 @@ new L<Net::Async::AMQP::Queue> instance once complete.
 =cut
 
 sub queue_declare {
-    my $self = shift;
-    my %args = @_;
-    die "No queue specified" unless defined $args{queue};
+	my $self = shift;
+	my %args = @_;
+	die "No queue specified" unless defined $args{queue};
 
-    $self->future->then(sub {
-        my $f = $self->loop->new_future;
-        $self->add_child(
+	$self->future->then(sub {
+		my $f = $self->loop->new_future;
+		$self->add_child(
 			my $q = Net::Async::AMQP::Queue->new(
 				amqp    => $self->amqp,
 				future  => $f,
@@ -187,35 +187,35 @@ sub queue_declare {
 		# triggering cleanup logic as soon as that channel
 		# goes out of scope.
 		Scalar::Util::weaken($q->{channel});
-        $self->debug_printf("Declaring queue [%s]", $args{queue});
-        my $frame = Net::AMQP::Frame::Method->new(
-            method_frame => Net::AMQP::Protocol::Queue::Declare->new(
-                queue       => Net::AMQP::Value::String->new($args{queue}),
-                passive     => $args{passive} || 0,
-                durable     => $args{durable} || 0,
-                exclusive   => $args{exclusive} || 0,
-                auto_delete => $args{auto_delete} || 0,
-                no_ack      => $args{no_ack} || 0,
+		$self->debug_printf("Declaring queue [%s]", $args{queue});
+		my $frame = Net::AMQP::Frame::Method->new(
+			method_frame => Net::AMQP::Protocol::Queue::Declare->new(
+				queue       => Net::AMQP::Value::String->new($args{queue}),
+				passive     => $args{passive} || 0,
+				durable     => $args{durable} || 0,
+				exclusive   => $args{exclusive} || 0,
+				auto_delete => $args{auto_delete} || 0,
+				no_ack      => $args{no_ack} || 0,
 				($args{arguments}
 				? (arguments   => $args{arguments})
 				: ()
 				),
-                ticket      => 0,
-                nowait      => 0,
-            )
-        );
-        $self->push_pending(
-            'Queue::DeclareOk' => sub {
-                my ($amqp, $frame) = @_;
-                my $method_frame = $frame->method_frame;
-                $q->queue_name($method_frame->queue);
-                $f->done($q) unless $f->is_ready;
-            }
-        );
+				ticket      => 0,
+				nowait      => 0,
+			)
+		);
+		$self->push_pending(
+			'Queue::DeclareOk' => sub {
+				my ($amqp, $frame) = @_;
+				my $method_frame = $frame->method_frame;
+				$q->queue_name($method_frame->queue);
+				$f->done($q) unless $f->is_ready;
+			}
+		);
 		$self->closure_protection($f);
-        $self->send_frame($frame);
-        $f;
-    })
+		$self->send_frame($frame);
+		$f;
+	})
 }
 
 =head2 publish
@@ -234,12 +234,12 @@ channel instance once the server has confirmed publishing is complete.
 =cut
 
 sub publish {
-    my $self = shift;
-    my %args = @_;
-    die "no exchange" unless exists $args{exchange};
+	my $self = shift;
+	my %args = @_;
+	die "no exchange" unless exists $args{exchange};
 
-    $self->future->then(sub {
-        my $f = $self->loop->new_future;
+	$self->future->then(sub {
+		my $f = $self->loop->new_future;
 		{ # When publishing a message, we should expect either an ACK, or a return.
 		  # Since these are mutually exclusive, we also need to remove the pending
 		  # handler for the opposing event once one event has been received. Note
@@ -258,11 +258,11 @@ sub publish {
 				my $method_frame = $frame->method_frame;
 				# $self->remove_pending('Basic::Ack' => $ack);
 				$f->fail(
-                    $method_frame->reply_text,
-                    code     => $method_frame->reply_code,
-                    exchange => $method_frame->exchange,
-                    rkey     => $method_frame->routing_key
-                ) unless $f->is_ready;
+					$method_frame->reply_text,
+					code     => $method_frame->reply_code,
+					exchange => $method_frame->exchange,
+					rkey     => $method_frame->routing_key
+				) unless $f->is_ready;
 				weaken $ack;
 			};
 			$self->push_pending(
@@ -273,37 +273,37 @@ sub publish {
 			);
 		}
 
-        my @frames = $self->amqp->split_payload(
-            $args{payload},
-            exchange         => Net::AMQP::Value::String->new($args{exchange}),
-            mandatory		 => $args{mandatory} // 0,
-            immediate        => 0,
-            (exists $args{routing_key} ? (routing_key => Net::AMQP::Value::String->new($args{routing_key})) : ()),
-            ticket           => 0,
-            content_type     => 'application/binary',
-            content_encoding => undef,
-            timestamp        => time,
-            type             => Net::AMQP::Value::String->new($args{type}),
-            user_id          => $self->amqp->user,
-            no_ack           => 0,
+		my @frames = $self->amqp->split_payload(
+			$args{payload},
+			exchange         => Net::AMQP::Value::String->new($args{exchange}),
+			mandatory		 => $args{mandatory} // 0,
+			immediate        => 0,
+			(exists $args{routing_key} ? (routing_key => Net::AMQP::Value::String->new($args{routing_key})) : ()),
+			ticket           => 0,
+			content_type     => 'application/binary',
+			content_encoding => undef,
+			timestamp        => time,
+			type             => Net::AMQP::Value::String->new($args{type}),
+			user_id          => $self->amqp->user,
+			no_ack           => 0,
 #            headers          => {
 #                type => $args{type},
 #            },
-            delivery_mode    => 1,
-            priority         => 1,
-            correlation_id   => undef,
-            expiration       => undef,
-            message_id       => undef,
-            app_id           => undef,
-            cluster_id       => undef,
-            weight           => 0,
-        );
+			delivery_mode    => 1,
+			priority         => 1,
+			correlation_id   => undef,
+			expiration       => undef,
+			message_id       => undef,
+			app_id           => undef,
+			cluster_id       => undef,
+			weight           => 0,
+		);
 		$self->closure_protection($f);
-        $self->send_frame(
-            $_,
-        ) for @frames;
-        $f
-    })
+		$self->send_frame(
+			$_,
+		) for @frames;
+		$f
+	})
 }
 
 =head2 qos
@@ -324,27 +324,27 @@ channel instance once the operation is complete.
 =cut
 
 sub qos {
-    my $self = shift;
-    my %args = @_;
+	my $self = shift;
+	my %args = @_;
 
-    $self->future->then(sub {
-        my $f = $self->loop->new_future;
-        my $channel = $self->id;
-        $self->push_pending(
-            'Basic::QosOk' => [ $f, $self ],
-        );
+	$self->future->then(sub {
+		my $f = $self->loop->new_future;
+		my $channel = $self->id;
+		$self->push_pending(
+			'Basic::QosOk' => [ $f, $self ],
+		);
 
-        my $frame = Net::AMQP::Frame::Method->new(
-            method_frame => Net::AMQP::Protocol::Basic::Qos->new(
-                nowait         => 0,
-                prefetch_count => $args{prefetch_count},
-                prefetch_size  => $args{prefetch_size} || 0,
-            )
-        );
+		my $frame = Net::AMQP::Frame::Method->new(
+			method_frame => Net::AMQP::Protocol::Basic::Qos->new(
+				nowait         => 0,
+				prefetch_count => $args{prefetch_count},
+				prefetch_size  => $args{prefetch_size} || 0,
+			)
+		);
 		$self->closure_protection($f);
-        $self->send_frame($frame);
-        $f
-    });
+		$self->send_frame($frame);
+		$f
+	});
 }
 
 =head2 ack
@@ -361,31 +361,31 @@ channel instance once the operation is complete.
 =cut
 
 sub ack {
-    my $self = shift;
-    my %args = @_;
+	my $self = shift;
+	my %args = @_;
 
-    my $id = $self->id;
-    $self->future->on_done(sub {
-        my $channel = $id;
-        my $frame = Net::AMQP::Frame::Method->new(
-            method_frame => Net::AMQP::Protocol::Basic::Ack->new(
-               # nowait      => 0,
+	my $id = $self->id;
+	$self->future->on_done(sub {
+		my $channel = $id;
+		my $frame = Net::AMQP::Frame::Method->new(
+			method_frame => Net::AMQP::Protocol::Basic::Ack->new(
+			   # nowait      => 0,
 				delivery_tag => $args{delivery_tag},
 				multiple     => $args{multiple} // 0,
-            )
-        );
-        $self->send_frame($frame);
-    });
+			)
+		);
+		$self->send_frame($frame);
+	});
 }
 
 =pod
 
 Example output:
 
-        'method_id' => 40,
-        'reply_code' => 404,
-        'class_id' => 60,
-        'reply_text' => 'NOT_FOUND - no exchange \'invalidchan\' in vhost \'vhost\''
+		'method_id' => 40,
+		'reply_code' => 404,
+		'class_id' => 60,
+		'reply_text' => 'NOT_FOUND - no exchange \'invalidchan\' in vhost \'vhost\''
 
 =cut
 
@@ -396,14 +396,14 @@ Called when the channel has been closed.
 =cut
 
 sub on_close {
-    my $self = shift;
-    my $frame = shift;
-    $self->bus->invoke_event(
-        'close',
-        code => $frame->reply_code,
-        message => $frame->reply_text,
-    );
-    $self->amqp->channel_closed($self->id);
+	my $self = shift;
+	my $frame = shift;
+	$self->bus->invoke_event(
+		'close',
+		code => $frame->reply_code,
+		message => $frame->reply_text,
+	);
+	$self->amqp->channel_closed($self->id);
 }
 
 =head2 send_frame
@@ -436,23 +436,23 @@ channel instance once the operation is complete.
 =cut
 
 sub close {
-    my $self = shift;
-    my %args = @_;
-    $self->debug_printf("Close channel %d", $self->id);
+	my $self = shift;
+	my %args = @_;
+	$self->debug_printf("Close channel %d", $self->id);
 
-    my $f = $self->loop->new_future;
-    my $frame = Net::AMQP::Frame::Method->new(
-        method_frame => Net::AMQP::Protocol::Channel::Close->new(
+	my $f = $self->loop->new_future;
+	my $frame = Net::AMQP::Frame::Method->new(
+		method_frame => Net::AMQP::Protocol::Channel::Close->new(
 			reply_code  => $args{code} // 404,
 			reply_text  => $args{text} // 'closing',
-        )
-    );
-    $self->push_pending(
-        'Channel::CloseOk' => [ $f, $self ],
-    );
+		)
+	);
+	$self->push_pending(
+		'Channel::CloseOk' => [ $f, $self ],
+	);
 	$self->closure_protection($f);
-    $self->send_frame($frame);
-    return $f;
+	$self->send_frame($frame);
+	return $f;
 }
 
 =head2 push_pending
@@ -460,12 +460,12 @@ sub close {
 =cut
 
 sub push_pending {
-    my $self = shift;
-    while(@_) {
-        my ($type, $code) = splice @_, 0, 2;
-        push @{$self->{pending}{$type}}, $code;
-    }
-    return $self;
+	my $self = shift;
+	while(@_) {
+		my ($type, $code) = splice @_, 0, 2;
+		push @{$self->{pending}{$type}}, $code;
+	}
+	return $self;
 }
 
 =head2 remove_pending
@@ -478,8 +478,8 @@ Returns C< $self >.
 
 sub remove_pending {
 	my $self = shift;
-    while(@_) {
-        my ($type, $code) = splice @_, 0, 2;
+	while(@_) {
+		my ($type, $code) = splice @_, 0, 2;
 		# This is the same as extract_by { $_ eq $code } @{$self->{pending}{$type}};,
 		# but since we'll be calling it a lot might as well do it inline:
 		splice
@@ -488,13 +488,13 @@ sub remove_pending {
 			1 for grep {
 				$self->{pending}{$type}[$_] eq $code
 			} reverse 0..$#{$self->{pending}{$type}};
-    }
-    return $self;
+	}
+	return $self;
 }
 
 =head2 next_pending
 
-Retrieves the next pending handler for the given incoming frame type (see L</get_frame_type>),
+Retrieves the next pending handler for the given incoming frame type (see L<Net::Async::AMQP::Utils/amqp_frame_type>),
 and calls it.
 
 Takes the following parameters:
@@ -510,75 +510,80 @@ Returns $self.
 =cut
 
 sub next_pending {
-    my ($self, $frame) = @_;
+	my ($self, $frame) = @_;
 
-    # First part of a frame. There's more to come, so stash a new future
-    # and return.
-    if($frame->isa('Net::AMQP::Frame::Header')) {
+	# First part of a frame. There's more to come, so stash a new future
+	# and return.
+	if($frame->isa('Net::AMQP::Frame::Header')) {
 		$self->{incoming_message}{type} = $frame->header_frame->type;
-        if($frame->header_frame->headers) {
-            eval {
+		if($frame->header_frame->headers) {
+			eval {
 				$self->{incoming_message}{type} = $frame->header_frame->headers->{type}
 					if exists $frame->header_frame->headers->{type};
 				1
 			} or $self->debug_printf("Unexpected exception while doing something: %s", $@);
-        }
+		}
 
 		# Messages may be empty - in this case we'd have no body frames at all, we're done already:
-        unless($frame->body_size) {
-            $self->{incoming_message}{payload} = '';
-            $self->bus->invoke_event(
-                message => @{$self->{incoming_message}}{qw(type payload ctag dtag rkey)},
-            );
-            delete $self->{incoming_message};
-        }
+		unless($frame->body_size) {
+			$self->{incoming_message}{payload} = '';
+			$self->bus->invoke_event(
+				message => @{$self->{incoming_message}}{qw(type payload ctag dtag rkey)},
+			);
+			delete $self->{incoming_message};
+		}
 
-        return $self;
-    }
+		return $self;
+	}
 
-    # Body part of an incoming message.
-    # TODO should handle multiple chunks?
-    if($frame->isa('Net::AMQP::Frame::Body')) {
-        $self->{incoming_message}{payload} = $frame->payload;
-        $self->bus->invoke_event(
-            message => @{$self->{incoming_message}}{qw(type payload ctag dtag rkey)},
-        );
-        delete $self->{incoming_message};
-        return $self;
-    }
+	# Body part of an incoming message.
+	# TODO should handle multiple chunks?
+	if($frame->isa('Net::AMQP::Frame::Body')) {
+		$self->{incoming_message}{payload} = $frame->payload;
+		$self->bus->invoke_event(
+			message => @{$self->{incoming_message}}{qw(type payload ctag dtag rkey)},
+		);
+		delete $self->{incoming_message};
+		return $self;
+	}
 
-    return $self unless $frame->can('method_frame') && (my $method_frame = $frame->method_frame);
-    my $type = $self->amqp->get_frame_type($frame);
+	return $self unless $frame->can('method_frame') && (my $method_frame = $frame->method_frame);
+	my $type = amqp_frame_type($frame);
 	if($type eq 'Basic::CancelOk') {
 		my ($ctag) = ($method_frame->consumer_tag);
-        $self->debug_printf("Cancel $ctag");
+		$self->debug_printf("Cancel $ctag");
 		$self->bus->invoke_event(
 			'cancel',
 			ctag => $ctag,
 		);
+		# Also raise this as a "listener_stop"
+		# event, for managed channels
+		$self->bus->invoke_event(
+			listener_stop => $ctag
+		);
 	}
 
 	# Message delivery, part 3: The "Deliver" message.
 	# This is actually where we start.
-    if($type eq 'Basic::Deliver') {
-        $self->debug_printf("Already have incoming_message?") if exists $self->{incoming_message};
-        $self->{incoming_message} = {
-            ctag => $method_frame->consumer_tag,
-            dtag => $method_frame->delivery_tag,
-            rkey => $method_frame->routing_key,
-        };
-        return $self;
-    }
+	if($type eq 'Basic::Deliver') {
+		$self->debug_printf("Already have incoming_message?") if exists $self->{incoming_message};
+		$self->{incoming_message} = {
+			ctag => $method_frame->consumer_tag,
+			dtag => $method_frame->delivery_tag,
+			rkey => $method_frame->routing_key,
+		};
+		return $self;
+	}
 
-    if($type eq 'Channel::Close') {
-        $self->debug_printf("Channel was %d, calling close", $frame->channel);
-        $self->on_close(
-            $method_frame
-        );
-        return $self;
-    }
+	if($type eq 'Channel::Close') {
+		$self->debug_printf("Channel was %d, calling close", $frame->channel);
+		$self->on_close(
+			$method_frame
+		);
+		return $self;
+	}
 
-    if(my $next = shift @{$self->{pending}{$type} || []}) {
+	if(my $next = shift @{$self->{pending}{$type} || []}) {
 		# We have a registered handler for this frame type. This usually
 		# means that we've sent a message and are awaiting a response.
 		if(ref($next) eq 'ARRAY') {
@@ -641,10 +646,10 @@ This channel ID.
 =cut
 
 sub id {
-    my $self = shift;
-    return $self->{id} unless @_;
-    $self->{id} = shift;
-    $self
+	my $self = shift;
+	return $self->{id} unless @_;
+	$self->{id} = shift;
+	$self
 }
 
 sub as_string {
